import pathlib

from click import Path, argument, option

option_background_file = option(
    "-b", "--background-file", "background_file",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
        path_type=pathlib.Path,
    ),
    required=True,
    help="SVG file to use as background.",
)

arguments_foreground_filename = argument(
    "foreground_filename",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        allow_dash=False,
        path_type=pathlib.Path,
    ),
    nargs=-1,
)

option_output_dir = option(
    "-o", "--output-dir", "output_dir",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        executable=True,
        allow_dash=False,
        path_type=pathlib.Path,
    ),
    required=True,
    help="Base directory where created emblems should be stored in. Must already exist.",
)

option_save_svg = option(
    "-s/-S", "--save-svg/--ignore-svg", "save_svg",
    type=bool,
    is_flag=True,
    default=True,
    help="Whether an emblem in SVG format should be generated or not.",
)

option_save_png = option(
    "-p/-P", "--save-png/--ignore-png", "save_png",
    type=bool,
    is_flag=True,
    default=True,
    help="Whether an emblem in PNG format should be generated or not.",
)

option_width_px = option(
    "-w", "--width-px", "width_px",
    type=float,
    default=512.0,
    help="Width in pixels of the output.",
)

option_height_px = option(
    "-h", "--height-px", "height_px",
    type=float,
    default=512.0,
    help="Height in pixels of the output.",
)

option_foreground_color = option(
    "-c", "--foreground-color", "foreground_color",
    type=str,
    required=True,
    help="CSS color (`red`, `#ff0000`, `rgb(255 0 0)`...) of the foreground.",
)

option_primary_color = option(
    "-c", "--primary-color", "primary_color",
    type=str,
    required=True,
    help="Primary CSS color (`red`, `#ff0000`, `rgb(255 0 0)`...) of the foreground.",
)

option_secondary_color = option(
    "-C", "--secondary-color", "secondary_color",
    type=str,
    required=True,
    help="Secondary CSS color (`lime`, `#00ff00`, `rgb(0 255 0)`...) of the foreground.",
)

option_shadow_offset_x_px = option(
    "-x", "--shadow-offset-x-px", "shadow_offset_x_px",
    type=float,
    required=True,
    help="Horizontal pixel offset of the shadow.",
)

option_shadow_offset_y_px = option(
    "-y", "--shadow-offset-y-px", "shadow_offset_y_px",
    type=float,
    required=True,
    help="Vertical pixel offset of the shadow.",
)

option_shadow_stddev = option(
    "-r", "--shadow-stddev", "shadow_stddev",
    type=float,
    required=True,
    help="Shadow blur standard deviation.",
)

option_shadow_color = option(
    "-f", "--shadow-color", "shadow_color",
    type=str,
    required=True,
    help="CSS color (`black`, `#000000`, `rgb(0 0 0)`...) of the shadow.",
)

__all__ = (
    "option_background_file",
    "arguments_foreground_filename",
    "option_output_dir",
    "option_save_svg",
    "option_save_png",
    "option_width_px",
    "option_height_px",
    "option_foreground_color",
    "option_primary_color",
    "option_secondary_color",
    "option_shadow_offset_x_px",
    "option_shadow_offset_y_px",
    "option_shadow_stddev",
    "option_shadow_color",
)



"""
Module containing functions to operate on files or directories.
"""

import subprocess
from pathlib import Path
from typing import Iterable


class WorkUnit:
    def __init__(self, source: Path, destination_stem: Path):
        self.source: Path = source
        self.destination_stem: Path = destination_stem

    @classmethod
    def find(cls, sources: Iterable[Path], destination: Path, dir_glob: str = "**/*.svg") -> set["WorkUnit"]:
        """
        Create a :class:`set` of :class:`.WorkUnit` from a list of sources and a destination directory.

        If a source is a directory, this function selects all files inside matching ``dir_glob``, producing destination paths preserving the original source directory structure.

        .. warning::
            This method **assumes the sources exist**. Check before calling it!

        .. warning::
            This method **assumes the destination exists and is a directory already**. Check before calling it!

        :param sources: An iterable of the paths to the sources to use. Sources can be both files or directories.
        :param destination: Path to the destination directory.
        :param dir_glob: Glob to use to find files inside directory sources.
        :return: The resulting :class:`set` of :class:`.WorkUnit`.
        """

        collection = set()

        for source in sources:
            if not source.exists():
                raise FileNotFoundError("Source does not exist: ", source)

            if source.is_dir():
                for file in source.glob(dir_glob):
                    # Skip directories
                    if file.is_dir():
                        continue

                    relative = file.relative_to(source)
                    destination_stem = destination / relative.parent / relative.stem
                    collection.add(cls(source=file, destination_stem=destination_stem))

            else:
                destination_stem = destination / source.stem
                collection.add(cls(source=source, destination_stem=destination_stem))

        return collection

    def create_destination_parents(self):
        """
        Create the parent directories of :attr:`.destination_stem`, if they do not already exist.
        """
        self.destination_stem.parent.mkdir(parents=True, exist_ok=True)

    def destination_svg(self) -> Path:
        """
        :return: :attr:`.destination_stem`, but with the ``.svg`` file extension.
        """
        return self.destination_stem.with_name(self.destination_stem.name + ".svg")

    def write_svg(self, svg_data: str) -> Path:
        """
        Write the given SVG document to :meth:`.destination_svg`.

        .. warning::
            This method **assumes the parent directories exist**. Create them with :meth:`.create_destination_parents` before running it!

        :param svg_data: The SVG document, in string form.
        """

        destination_svg_path = self.destination_svg()

        with destination_svg_path.open("w", encoding="utf-8") as destination_svg_file:
            destination_svg_file.write(svg_data)

        return destination_svg_path

    def destination_png(self) -> Path:
        """
        :return: :attr:`.destination_stem`, but with the ``.png`` file extension.
        """
        return self.destination_stem.with_name(self.destination_stem.name + ".png")

    def write_png(self, svg_data: str, *, width_px: int, height_px: int) -> Path:
        """
        Write the given SVG document as PNG to :meth:`.destination_png`, calling Inkscape in a :mod:`subprocess` to perform the conversion.

        .. warning::
            This method **assumes the parent directories exist**. Create them with :meth:`.create_destination_parents` before running it!

        :param svg_data: The SVG document, in string form.
        :param width_px: The width in pixels of the resulting image.
        :param height_px: The height in pixels of the resulting image.
        """

        destination_png_path = self.destination_png()

        result = subprocess.run(
            args=[
                "inkscape",
                "--pipe",
                "--export-type=png",
                f"--export-filename={self.destination_png()}",
                f"--export-width={width_px}",
                f"--export-height={height_px}"
            ],
            input=svg_data.encode("utf-8"),
        )
        if result.returncode != 0:
            raise ChildProcessError("Conversion to PNG with Inkscape returned non-zero exit code: ", result.returncode)

        return destination_png_path


__all__ = (
    "WorkUnit",
)



import pathlib
from typing import Iterator

import bs4
import click

from emblematic.composition import compose_basic, compose_solid, compose_shadow, compose_duotone
from .argopts import option_background_file, arguments_foreground_filename, option_output_dir, option_save_svg, option_save_png, option_width_px, \
    option_height_px, option_foreground_color, option_primary_color, option_secondary_color, option_shadow_offset_x_px, option_shadow_offset_y_px, \
    option_shadow_stddev, option_shadow_color
from .files import WorkUnit


@click.group()
@click.version_option(package_name="emblematic", prog_name="emblematic")
@click.pass_context
def main(ctx: click.Context):
    """
    Generate an emblem, an easily recognizable yet consistently styled icon.
    """

    # 80 is NOT a sensible default in any case, let the user choose!
    ctx.max_content_width = 10000
    ctx.show_default = True


@main.command()
@option_background_file
@arguments_foreground_filename
@option_output_dir
@option_save_svg
@option_save_png
@option_width_px
@option_height_px
def basic(
    background_file: pathlib.Path,
    foreground_filename: Iterator[pathlib.Path],
    output_dir: pathlib.Path,
    save_svg: bool,
    save_png: bool,
    width_px: float,
    height_px: float,
):
    """
    Create a simple emblem.
    """

    click.echo(
        err=True,
        message=
        click.style("Generating ", fg="cyan") +
        click.style("basic", fg="cyan", bold=True) +
        click.style(" emblems.", fg="cyan")
    )

    work_units = WorkUnit.find(sources=foreground_filename, destination=output_dir)

    click.echo(
        err=True,
        message=
        click.style("Found ", fg="cyan") +
        click.style(f"{len(work_units)}", fg="cyan", bold=True) +
        click.style(" foreground elements.", fg="cyan")
    )

    with background_file.open("r") as background_file:
        background_soup_original = bs4.BeautifulSoup(background_file, features="lxml-xml")

    for work_unit in work_units:
        background_soup = background_soup_original.__copy__()

        with work_unit.source.open("r") as foreground_file:
            foreground_soup = bs4.BeautifulSoup(foreground_file, features="lxml-xml")

        result = compose_basic(
            background=background_soup,
            foreground=foreground_soup,
            final_width_px=width_px,
            final_height_px=height_px,
        ).prettify()

        work_unit.create_destination_parents()

        if save_svg:
            svg_path = work_unit.write_svg(result)
            click.echo(svg_path)

        if save_png:
            png_path = work_unit.write_png(result, width_px=int(width_px), height_px=int(height_px))
            click.echo(png_path)


@main.command()
@option_background_file
@arguments_foreground_filename
@option_output_dir
@option_save_svg
@option_save_png
@option_width_px
@option_height_px
@option_foreground_color
def solid(
    background_file: pathlib.Path,
    foreground_filename: Iterator[pathlib.Path],
    output_dir: pathlib.Path,
    save_svg: bool,
    save_png: bool,
    width_px: float,
    height_px: float,
    foreground_color: str,
):
    """
    Create an emblem with a solid color foreground.

    Intended for Font Awesome Solid icons.
    """

    click.echo(
        err=True,
        message=
        click.style("Generating ", fg="cyan") +
        click.style("solid", fg="cyan", bold=True) +
        click.style(" emblems.", fg="cyan")
    )

    work_units = WorkUnit.find(sources=foreground_filename, destination=output_dir)

    click.echo(
        err=True,
        message=
        click.style("Found ", fg="cyan") +
        click.style(f"{len(work_units)}", fg="cyan", bold=True) +
        click.style(" foreground elements.", fg="cyan")
    )

    with background_file.open("r") as background_file:
        background_soup_original = bs4.BeautifulSoup(background_file, features="lxml-xml")

    for work_unit in work_units:
        background_soup = background_soup_original.__copy__()

        with work_unit.source.open("r") as foreground_file:
            foreground_soup = bs4.BeautifulSoup(foreground_file, features="lxml-xml")

        result = compose_solid(
            background=background_soup,
            foreground=foreground_soup,
            foreground_fill=foreground_color,
            final_width_px=width_px,
            final_height_px=height_px,
        ).prettify()

        work_unit.create_destination_parents()

        if save_svg:
            svg_path = work_unit.write_svg(result)
            click.echo(svg_path)

        if save_png:
            png_path = work_unit.write_png(result, width_px=int(width_px), height_px=int(height_px))
            click.echo(png_path)


@main.command()
@option_background_file
@arguments_foreground_filename
@option_output_dir
@option_save_svg
@option_save_png
@option_width_px
@option_height_px
@option_foreground_color
@option_shadow_offset_x_px
@option_shadow_offset_y_px
@option_shadow_stddev
@option_shadow_color
def shadow(
    background_file: pathlib.Path,
    foreground_filename: Iterator[pathlib.Path],
    output_dir: pathlib.Path,
    save_svg: bool,
    save_png: bool,
    width_px: float,
    height_px: float,
    foreground_color: str,
    shadow_offset_x_px: float,
    shadow_offset_y_px: float,
    shadow_stddev,
    shadow_color,
):
    """
    Create an emblem with a solid color foreground with a shadow.

    Intended for Font Awesome Solid icons.
    """

    click.echo(
        err=True,
        message=
        click.style("Generating ", fg="cyan") +
        click.style("shadow", fg="cyan", bold=True) +
        click.style(" emblems.", fg="cyan")
    )

    work_units = WorkUnit.find(sources=foreground_filename, destination=output_dir)

    click.echo(
        err=True,
        message=
        click.style("Found ", fg="cyan") +
        click.style(f"{len(work_units)}", fg="cyan", bold=True) +
        click.style(" foreground elements.", fg="cyan")
    )

    with background_file.open("r") as background_file:
        background_soup_original = bs4.BeautifulSoup(background_file, features="lxml-xml")

    for work_unit in work_units:
        background_soup = background_soup_original.__copy__()

        with work_unit.source.open("r") as foreground_file:
            foreground_soup = bs4.BeautifulSoup(foreground_file, features="lxml-xml")

        result = compose_shadow(
            background=background_soup,
            foreground=foreground_soup,
            foreground_fill=foreground_color,
            shadow_x_px=shadow_offset_x_px,
            shadow_y_px=shadow_offset_y_px,
            shadow_stddev=shadow_stddev,
            shadow_color=shadow_color,
            final_width_px=width_px,
            final_height_px=height_px,
        ).prettify()

        work_unit.create_destination_parents()

        if save_svg:
            svg_path = work_unit.write_svg(result)
            click.echo(svg_path)

        if save_png:
            png_path = work_unit.write_png(result, width_px=int(width_px), height_px=int(height_px))
            click.echo(png_path)


@main.command()
@option_background_file
@arguments_foreground_filename
@option_output_dir
@option_save_svg
@option_save_png
@option_width_px
@option_height_px
@option_primary_color
@option_secondary_color
def duotone(
    background_file: pathlib.Path,
    foreground_filename: Iterator[pathlib.Path],
    output_dir: pathlib.Path,
    save_svg: bool,
    save_png: bool,
    width_px: float,
    height_px: float,
    primary_color: str,
    secondary_color: str,
):
    """
    Create an emblem with two separate colors as foreground.

    Intended for Font Awesome Duotone icons.
    """

    click.echo(
        err=True,
        message=
        click.style("Generating ", fg="cyan") +
        click.style("duotone", fg="cyan", bold=True) +
        click.style(" emblems.", fg="cyan")
    )

    work_units = WorkUnit.find(sources=foreground_filename, destination=output_dir)

    click.echo(
        err=True,
        message=
        click.style("Found ", fg="cyan") +
        click.style(f"{len(work_units)}", fg="cyan", bold=True) +
        click.style(" foreground elements.", fg="cyan")
    )

    with background_file.open("r") as background_file:
        background_soup_original = bs4.BeautifulSoup(background_file, features="lxml-xml")

    for work_unit in work_units:
        background_soup = background_soup_original.__copy__()

        with work_unit.source.open("r") as foreground_file:
            foreground_soup = bs4.BeautifulSoup(foreground_file, features="lxml-xml")

        result = compose_duotone(
            background=background_soup,
            foreground=foreground_soup,
            primary_fill=primary_color,
            secondary_fill=secondary_color,
            final_width_px=width_px,
            final_height_px=height_px,
        ).prettify()

        work_unit.create_destination_parents()

        if save_svg:
            svg_path = work_unit.write_svg(result)
            click.echo(svg_path)

        if save_png:
            png_path = work_unit.write_png(result, width_px=int(width_px), height_px=int(height_px))
            click.echo(png_path)


__all__ = (
    "main",
)
